from functools import lru_cache
import numpy as np

from igrf_gh import igrf_gh


@lru_cache
def dipole(time):
    # DIPOLE Calculates the Earth's dipole lat/lng at the specified time
    #   This function is intended for internal use by the GEO2MAG and MAG2GEO
    #   functions.
    #
    #   Converting to geomagnetic coordinates (MAG) requires knowledge
    #   of the Earth's dipole vector (in MAG coordinates, the Z-axis is
    #   parallel to the Earth's dipole).
    #
    #   In theory, we want the latitude and longitude of the dipole
    #   axis; in reality, however, it is much quicker and easier to
    #   calculate the sine and cosine of these values.
    #
    #------------------------------------------------------------------------------%
    #
    #   Version 1.0  Date: 2016/02/05
    #
    #   Copyright 2016 Aaron Hendry <aaron@hendry.co.nz>
    #   Otago University Space Physics Department, Dunedin, New Zealand
    #       <http://www.physics.otago.ac.nz/space>
    #   Released under the Simplified BSD License:
    #       <http://opensource.org/licenses/BSD-2-Clause>
    #
    #------------------------------------------------------------------------------%

    g, h = igrf_gh(time)

    sq = g[1] * g[1] + h[1] * h[1]
    sqq = np.sqrt(sq)
    sqr = np.sqrt(sq + g[0] * g[0])

    clat = -g[0] / sqr
    slat = sqq / sqr
    clon = -g[1] / sqq
    slon = -h[1] / sqq
    return clat, slat, clon, slon
