import numpy as np

from cart2sph import cart2sph
from sph2cart import sph2cart
from geo2mag import geo2mag
from step import step


def fieldline(rh, lat, lon, time, maxiter=1000):
    # FIELDLINE traces the IGRF field to the dipole magnetic equator
    #   [R,MLAT,MLON] = FIELDLINE(RH,LAT,LON,TIME) traces the IGRF field
    #   from the specified start point to the dipole geomagnetic equator. RH gives
    #   the starting altitude (in RE), while LAT and LON give the starting
    #   geographic latitude and longitude (in degrees). RH, LAT, and LON can be
    #   scalars or COLUMN vectors (not row vectors), but must be the same size. TIME
    #   is a MATLAB DATENUM variable specifying the epoch to calculate the IGRF
    #   field line for.
    #
    #   FIELDLINE(...,maxiter) As above, but forces the code to stop after
    #   MAXITER iterations. By default this value is set to 1000, which should allow
    #   almost all field lines to be traced to completion. Setting this to values
    #   lower than ~100 should speed up code execution, as the expense of the
    #   accuracy of the final result.
    #
    #   This code is based on the GEOCOR subroutine in the GEO-CGM FORTRAN code.
    #
    # ------------------------------------------------------------------------------%
    #
    #   Version 1.0  Date: 2016/02/05
    #
    #   Copyright 2016 Aaron Hendry <aaron@hendry.co.nz>
    #   Otago University Space Physics Department, Dunedin, New Zealand
    #       <http://www.physics.otago.ac.nz/space>
    #   Released under the Simplified BSD License:
    #       <http://opensource.org/licenses/BSD-2-Clause>
    #
    # ------------------------------------------------------------------------------%

    # Convert to radians
    lat = np.radians(lat)
    lon = np.radians(lon)

    # tolerance for stopping point
    err = 0.0001571

    # Convert to spherical magnetic units
    x, y, z = sph2cart(rh, np.pi / 2 - lat, lon)
    # Convert to MAG coordinates
    xm, ym, zm = geo2mag(x, y, z, time)
    # Convert back to spherical units
    ltm = np.arctan2(np.sqrt(xm**2 + ym**2), zm)

    # Distance to the (dipole) magnetic equator
    rl = rh / np.sin(ltm)**2
    zm_start = zm  # tells us the starting equator

    frac = 0.03 / (1 + 3 / (rl - 0.6))
    frac[zm_start < 0] = -frac[zm_start < 0]

    ds = rh * frac

    x1 = x.copy()
    y1 = y.copy()
    z1 = z.copy()

    i = 1
    filt = np.full_like(x, True, dtype=bool)

    while np.any(filt):
        x[filt], y[filt], z[filt] = step(x[filt], y[filt], z[filt], ds[filt], time)
        xm, ym, zm = geo2mag(x, y, z, time)
        rm = xm**2 + ym**2
        ltm = np.arctan2(np.sqrt(rm), zm)
        rm = np.sqrt(rm + zm**2)

        # If we've traced too long - we're probably not going to find a result
        if i >= maxiter:
            break

        filt_rm = np.isnan(x) | (rm < rh)
        x[filt_rm] = np.nan
        y[filt_rm] = np.nan
        z[filt_rm] = np.nan
        filt = ~filt_rm & ~(np.abs(np.pi / 2 - ltm) < err)
        filt_retry = filt & (np.sign(zm) != np.sign(zm_start))
        filt_cont = filt & (np.sign(zm) == np.sign(zm_start))
        x[filt_retry] = x1[filt_retry]
        y[filt_retry] = y1[filt_retry]
        z[filt_retry] = z1[filt_retry]
        ds[filt_retry] = ds[filt_retry] / 2
        x1[filt_cont] = x[filt_cont]
        y1[filt_cont] = y[filt_cont]
        z1[filt_cont] = z[filt_cont]
        ds[filt_cont] = rm[filt_cont] * frac[filt_cont]
        i = i + 1

    xm, ym, zm = geo2mag(x, y, z, time)
    r, lat, lon = cart2sph(xm, ym, zm)

    st = np.sin(lat)
    lat = np.pi / 2 - np.arctan(st * np.sqrt(np.abs(rh / (r - rh * st * st))))
    lat[zm_start < 0] = -lat[zm_start < 0]
    sn = np.sin(np.pi / 2 - lat)

    r = rh / (sn * sn)
    # Convert back to degrees
    mlat = np.degrees(lat)
    mlon = np.degrees(lon)

    return r, mlat, mlon
