import numpy as np
from scipy.io import loadmat

from igrf_gh import igrf_gh
from igrf_pq import igrf_pq


def numel(x):
    return np.prod(x.shape)


def igrf(rr, tt, ff, time):
    # IGRF Calculates IGRF magnetic field components
    #   [BX,BY,BZ] = IGRF(R,T,F,TIME) Calculates the cartesian magnetic field
    #   components for the location(s) <R,T,F>, where R is the radial distance from
    #   the center of the Earth, in units of RE, T is the geographic co-latitude in
    #   radians, and F is the geographic longitude in radians. R, T, and F can be
    #   scalars, vectors, or matrices, but they must be the same size. TIME gives
    #   the epoch to calculate IGRF for, and must be a scalar MATLAB DATENUM
    #   variable.
    #
    #   [...,BR,BT,BP] = IGRF(...) Calculates IGRF as above, but also provides the
    #   spherical magnetic field components.
    #
    #   IMPORTANT NOTE:
    #
    #   Due to how this function calculates the IGRF magnetic field components,
    #   particularly large input matrices may cause significant memory usage. To
    #   prevent potential out of memory errors and/or computer freezes, this
    #   function has been limited to computing 300,000 IGRF components
    #   simultaenously. If a greater number of components than this are needed, they
    #   should be split up into batches.
    #
    #   Also note that, although this function is called IGRF.m, it only uses the
    #   IGRF field model for the years 1900-2025. For years outside this range,
    #   other field models are used.
    #
    # ------------------------------------------------------------------------------%
    #
    #   Version 1.0  Date: 2016/02/05
    #
    #   Copyright 2016 Aaron Hendry <aaron@hendry.co.nz>
    #   Otago University Space Physics Department, Dunedin, New Zealand
    #       <http://www.physics.otago.ac.nz/space>
    #   Released under the Simplified BSD License:
    #       <http://opensource.org/licenses/BSD-2-Clause>
    #
    # ------------------------------------------------------------------------------%

    # COMMENT OUT THESE LINES TO REMOVE MEMORY CHECKS
    if numel(rr) > 300000:
        print(
            "geo2cgm:igrf",
            "The passed matrix is too large, and may cause "
            "memory issues. Try again with a smaller matrix, or subdivide the "
            "matrix and make multiple runs",
        )
        return None

    if not hasattr(igrf, "o"):
        data = loadmat("igrf_coeff.mat")
        igrf.gn = data["gn"].astype(int)[:, 0]
        igrf.gm = data["gm"].astype(int)[:, 0]
        igrf.g, igrf.h = igrf_gh(time)
        igrf.curr_time = time
        acc = 0.01
        th = np.radians(np.arange(0, 360 + acc, acc))
        igrf.acc = np.radians(acc)
        igrf.p, igrf.q = igrf_pq(th)
        igrf.o = np.ones((104,), dtype=int)

    if igrf.curr_time != time:
        igrf.g, igrf.h = igrf_gh(time)
        igrf.curr_time = time

    sz = tt.shape
    r = rr.T
    t = tt.T
    f = ff.T

    # Generate the P and Q legendre function coefficients.

    interp = t / igrf.acc
    ind = np.round(t / igrf.acc).astype(int)
    interp = interp[None, :] - ind
    interp = interp[igrf.o - 1]
    P = igrf.p[:, ind] * (1 - interp) + igrf.p[:, ind + 1] * interp
    Q = igrf.q[:, ind] * (1 - interp) + igrf.q[:, ind + 1] * interp
    # [P,Q] = igrf_pq_quicker(t);

    gmf = np.arange(14)[:, None] * f
    cf = np.cos(gmf)
    cf = cf[igrf.gm, :]
    sf = np.sin(gmf)
    sf = sf[igrf.gm, :]
    st = np.sin(t)
    # Calculate bx, by, and bz by taking the derivative of V
    rgn = np.cumprod(1.0 / np.tile(r, (15, 1)), axis=0)
    rgn = rgn[igrf.gn + 1, :]
    gchs = igrf.g[:, None] * cf + igrf.h[:, None] * sf
    bt = -np.sum(rgn * gchs * Q, axis=0)
    bpp = -(igrf.gm.T * igrf.g)[:, None] * sf + (igrf.gm.T * igrf.h)[:, None] * cf
    bp = -np.sum(rgn * bpp * P, axis=0)
    bp[bp != 0] = bp[bp != 0] / st[bp != 0]
    br = np.sum(((igrf.gn + 1)[:, None] * rgn) * gchs * P, axis=0)

    ct = np.cos(t)
    sp = np.sin(f)
    cp = np.cos(f)

    be = br * st + bt * ct

    bx = np.reshape(be * cp - bp * sp, sz)
    by = np.reshape(be * sp + bp * cp, sz)
    bz = np.reshape(br * ct - bt * st, sz)

    br = np.reshape(br, sz)
    bt = np.reshape(bt, sz)
    bp = np.reshape(bp, sz)

    return bx, by, bz, br, bt, bp
