from datetime import datetime
from scipy.io import loadmat


def igrf_gh(time):
    # IGRF_GH calculates the IGRF coefficients for the passed date
    #   [GT,HT] = IGRF_GH(time) generates the g and h coefficients for IGRF
    #   calculation.
    #
    # IGRF coefficients are calculated based on the current year (month and day can
    # be used as well, if desired). The current IGRF coefficients are stored in the
    # file igrf_coeff as variables G and H. G and H are both stored as [104xY]
    # matrices, where Y is a variable that will change based on the current IGRF
    # iteration.
    #
    # As of the time of writing, IGRF coefficients go back to 1900, and exist for
    # every 5 years until 2015 (which means there are currently 24 sets of IGRF
    # coefficients) - there are also a set of secular variation (SV) coefficients,
    # which attempt to predict the changes for the next 5 years, and allow for dates
    # between 2015-2020 to be used as well. This means that as of 2016, G and H are
    # both 104 x 25 matrices.
    #
    # New sets of IGRF coefficients can easily be added to the G and H matrices, by
    # appending a new column to the matrix. New coefficients are published on the
    # internet in many different locations (I used the following source:
    # http://www.ngdc.noaa.gov/IAGA/vmod/igrf12coeffs.txt). It is important that the
    # order of the coefficients is kept the same though. Each coefficient has two
    # indexes - n and m. The n-m order of the coefficients in G and H is the same,
    # and follows this pattern:
    #
    #   n | 1  1  2  2  2  3  3  3  3 ...
    #   m | 0  1  0  1  2  0  1  2  3 ...
    #
    # This order must be kept the same for each column, or everything will break.
    #
    # Note that when adding new columns to the G and H matrices, the last column
    # must ALWAYS be the SV coefficients. If for whatever reason SV coefficients
    # have not been released, the last columns should be filled with zeros.
    #
    #------------------------------------------------------------------------------%
    #
    #   Version 1.0  Date: 2016/02/05
    #
    #   Copyright 2016 Aaron Hendry <aaron@hendry.co.nz>
    #   Otago University Space Physics Department, Dunedin, New Zealand
    #       <http://www.physics.otago.ac.nz/space>
    #   Released under the Simplified BSD License:
    #       <http://opensource.org/licenses/BSD-2-Clause>
    #
    #------------------------------------------------------------------------------%

    if getattr(igrf_gh, "curr_time", None) == time:
        return igrf_gh.g, igrf_gh.h
    igrf_gh.curr_time = time

    if not hasattr(igrf_gh, "warned_hi"):
        igrf_gh.warned_lo = False
        igrf_gh.warned_hi = False

    data = loadmat("igrf_coeff.mat")
    G = data["G"]
    H = data["H"]
    years = data["years"][0].astype(int)

    # For non-epoch years (for instance 2012), coefficients can be calculated
    # by interpolating between the nearest two sets of coefficients, or using
    # the SV coefficients for years greater than the greatest current epoch.

    # We interpolate based on the MATLAB datenum value, as opposed to the year
    # (which is used in other CGM code).
    yrt = time.timetuple()
    yr = yrt[0]

    # If we occur before the earliest IGRF coefficients, throw an error, because we
    # have no way of extrapolating back that far.
    if yr < years[0]:
        print('geocgm:igrf_gh:invalid_epoch', 'Requested year (%d) is'
              ' outside the current IGRF coefficient range (1900-2025), '
              'and is beyond the scope of the GUFM 1 model (1590-1900). At present '
              'extrapolating backwards in time from %d is not supported.'
              % (yr, years[1]))

    # If we're outside of the current IGRF epoch, we're using Julien Aubert's
    # geodynamo-derived magnetic field model to predict the magnetic field. It's
    # more accurate than using secular variation.
    if yr > 2025:
        if yr <= years[-1]:
            if not igrf_gh.warned_hi:
                print('geocgm:igrf_gh:epoch_warn', 'Requested year (%d) is'
                    ' outside the current IGRF coefficient range (1900-2025). '
                    ' Secular variation'
                    ' values will be used to extrapolate to the required date.' % yr)
                igrf_gh.warned_hi = True
        else:
            if not igrf_gh.warned_hi:
                print('geocgm:igrf_gh:epoch_warn', 'Requested year (%d) is'
                    ' outside the current IGRF coefficient range (1900-2025). '
                    ' Secular variation'
                    ' values will be used to extrapolate to the required date.' % yr)
                igrf_gh.warned_hi = True
            tyr = yr - years[-1]
            ydiff = datetime(yr + 1, 1, 1) - datetime(yr, 1, 1)
            interp = tyr + (time - datetime(yr, 1, 1)) / ydiff
            igrf_gh.g = G[:, -2] + G[:, -1] * interp
            igrf_gh.h = H[:, -2] + H[:, -1] * interp
            return igrf_gh.g, igrf_gh.h
    if yr < 1900:
        if not igrf_gh.warned_lo:
            print('geocgm:igrf_gh:epoch_warn', 'Requested year (%d) is'
            ' outside the current IGRF coefficient range (1900-2025). '
            ' The GUFM 1 model will be used instead, though accuracy may be'
            ' decreased.' % yr)
            igrf_gh.warned_lo = True

    ind = years.searchsorted(yr)
    tyr = years[ind]

    ydiff = datetime(years[ind + 1], 1, 1) - datetime(tyr, 1, 1)
    f2 = (time - datetime(tyr, 1, 1)) / ydiff
    f1 = 1 - f2
    igrf_gh.g = G[:, ind] * f1 + G[:, ind + 1] * f2
    igrf_gh.h = H[:, ind] * f1 + H[:, ind + 1] * f2

    return igrf_gh.g, igrf_gh.h
