import numpy as np
from scipy.special import factorial


def numel(x):
    return np.prod(x.shape)


def igrf_pq(t):
    # IGRF_PQ Calculates the P and Q legendre polynomials for the IGRF calculations
    #   [P,Q] = IGRF_PQ(T) Calculates the Schmidt quasi-normalized associated
    #   Legendre functions of degree 1-13 and order 0-13, for the co-latitudes T. T
    #   can be scalar, vector, or matrix. P and Q will both be of size <104xN>,
    #   where N = numel(T), where P is the Legendre functions P(cos(T)), and Q is
    #   the first derivative of P: Q = dP/dT.
    #
    #   The n-m order of the resulting P and Q matrices is the same as the order of
    #   the coefficients from IGRF_GH; namely:
    #
    #       n | 1  1  2  2  2  3  3  3  3 ...
    #       m | 0  1  0  1  2  0  1  2  3 ...
    #
    #------------------------------------------------------------------------------%
    #
    #   Version 1.0  Date: 2016/02/05
    #
    #   Copyright 2016 Aaron Hendry <aaron@hendry.co.nz>
    #   Otago University Space Physics Department, Dunedin, New Zealand
    #       <http://www.physics.otago.ac.nz/space>
    #   Released under the Simplified BSD License:
    #       <http://opensource.org/licenses/BSD-2-Clause>
    #
    #------------------------------------------------------------------------------%

    t = np.asarray(t)

    ct = np.cos(t)
    st = np.sin(t)

    # The normalisation values are slightly expensive to calculate, so we
    # pre-calculate them to avoid having to do multiple calculations of them.
    # persistent norm ind
    if not hasattr(igrf_pq, "norm"):
        igrf_pq.ind = np.asarray([
            15,2,28,16,3,40,29,17,4,51,41,30,18,5,61,52,42,31,19,6,70,62,53,
            43,32,20,7,78,71,63,54,44,33,21,8,85,79,72,64,55,45,34,22,9,91,86,
            80,73,65,56,46,35,23,10,96,92,87,81,74,66,57,47,36,24,11,100,97,93,
            88,82,75,67,58,48,37,25,12,103,101,98,94,89,83,76,68,59,49,38,26,13,
            105,104,102,99,95,90,84,77,69,60,50,39,27,14,
        ])
        n = np.arange(1, 15)
        n = np.tile(n, (14, 1)).T
        m = np.flipud(np.tril(np.flipud(n))).T
        n = np.tril(n).T
        m = m[m.astype(bool)] - 1
        n = n[n.astype(bool)] - 1

        igrf_pq.norm = np.sqrt(2 * factorial(n - m) / factorial(n + m)) * (-1)**m
        igrf_pq.norm[m == 0] = 1

    nn = np.arange(1, 14)
    n21 = np.arange(1, 26, 2)
    n21c = (n21[:, None] * ct).T
    n21s = (n21[:, None] * st).T

    # P can be calculated recursively - MATLAB is not good at recursion, so I have
    # done the recursion manually.
    P = np.ones((numel(t), 105))
    P[:, 1:14] = np.cumprod(-n21s, axis=1)
    P[:, 14:27] = n21c * P[:, 0:13]
    P[:, 27:39] = 0.5 * (n21c[:, 1:] * P[:, 14:26] - np.arange(1, 24, 2) * P[:, 0:12])
    P[:, 39:50] = (n21c[:, 2:] * P[:, 27:38] - np.arange(2, 23, 2) * P[:, 14:25]) / 3
    P[:, 50:60] = 0.25 * (n21c[:, 3:] * P[:, 39:49] - np.arange(3, 22, 2) * P[:, 27:37])
    P[:, 60:69] = 0.2*(n21c[:, 4:]*P[:,50:59] - np.arange(4, 21, 2)*P[:,39:48])
    P[:, 69:77] = (n21c[:, 5:]*P[:,60:68] - np.arange(5, 20, 2)*P[:,50:58])/6
    P[:, 77:84] = (n21c[:, 6:]*P[:,69:76] - np.arange(6, 19, 2)*P[:,60:67])/7
    P[:, 84:90] = 0.125*(n21c[:,7:]*P[:,77:83] - np.arange(7, 18, 2)*P[:,69:75])
    P[:, 90:95] = (n21c[:, 8:]*P[:,84:89] - np.arange(8, 17, 2)*P[:,77:82])/9
    P[:, 95:99] = 0.1*(n21c[:, 9:]*P[:,90:94] - np.arange(9, 16, 2)*P[:,84:88])
    P[:, 99:102] = (n21c[:, 10:]*P[:,95:98] - np.arange(10, 15, 2)*P[:,90:93])/11
    P[:, 102:104] = (n21c[:, 11:]*P[:,99:101] - np.arange(11, 14, 2)*P[:,95:97])/12
    P[:, 104] = (25 * P[:, 102] * ct - 12 * P[:, 99]) / 13

    # Calculating Q, similar to P.
    Q = np.zeros((numel(t), 105))
    Q[:, 1:14] = -nn * P[:, 14:27]
    Q[:, 14:27] = (n21c * Q[:, 0:13] - n21s * P[:, 0:13]);
    Q[:, 27:39] = 0.5 * (n21c[:, 1:] * Q[:, 14:26] - n21s[:, 1:] * P[:, 14:26] - np.arange(1, 24, 2) * Q[:, 0:12])
    Q[:, 39:50] = (n21c[:, 2:] * Q[:, 27:38] - n21s[:, 2:]*P[:, 27:38] - np.arange(2, 23, 2)*Q[:, 14:25]) / 3
    Q[:, 50:60] = 0.25 * (n21c[:, 3:]*Q[:, 39:49] - n21s[:, 3:]*P[:, 39:49] - np.arange(3, 22, 2)*Q[:, 27:37])
    Q[:, 60:69] = 0.2*(n21c[:, 4:]*Q[:, 50:59] - n21s[:,4:]*P[:,50:59] - np.arange(4, 21, 2)*Q[:, 39:48])
    Q[:, 69:77] = (n21c[:, 5:]*Q[:, 60:68] - n21s[:,5:]*P[:, 60:68] - np.arange(5, 20, 2)*Q[:, 50:58])/6
    Q[:, 77:84] = (n21c[:, 6:]*Q[:, 69:76] - n21s[:,6:]*P[:, 69:76] - np.arange(6, 19, 2)*Q[:, 60:67])/7
    Q[:, 84:90] = 0.125*(n21c[:, 7:]*Q[:, 77:83] - n21s[:,7:]*P[:, 77:83] - np.arange(7, 18, 2)*Q[:, 69:75])
    Q[:, 90:95] = (n21c[:, 8:]*Q[:, 84:89] - n21s[:,8:]*P[:, 84:89] - np.arange(8, 17, 2)*Q[:, 77:82])/9
    Q[:, 95:99] = 0.1*(n21c[:, 9:]*Q[:, 90:94] - n21s[:,9:]*P[:, 90:94] - np.arange(9, 16, 2)*Q[:, 84:88])
    Q[:, 99:102] = (n21c[:, 10:]*Q[:, 95:98] - n21s[:,10:]*P[:, 95:98] - np.arange(10, 15, 2)*Q[:, 90:93])/11
    Q[:, 102:104] = (n21c[:, 11:]*Q[:, 99:101] - n21s[:,11:]*P[:, 99:101] - np.arange(11, 14, 2)*Q[:, 95:97])/12
    Q[:, 104] = (25 * Q[:, 102] * ct - 12 * Q[:, 99]) / 13

    # Normalise P by the Schmidt normalisation procedure.
    P = (P * igrf_pq.norm).T
    Q = (Q * igrf_pq.norm).T
    # Reorder P and Q to be in the right order
    P = P[igrf_pq.ind - 1, :]
    Q = Q[igrf_pq.ind - 1, :]

    return P, Q
